@@ -13,13 +13,14 @@ class MarkdownExporter(TemplateExporter):
     """
     Exports to a markdown document (.md)
     """
+    export_from_notebook = "Markdown"
 
     @default('file_extension')
     def _file_extension_default(self):
         return '.md'
 
-    @default('template_file')
-    def _template_file_default(self):
+    @default('template_name')
+    def _template_name_default(self):
         return 'markdown'
 
     output_mimetype = 'text/markdown'
@@ -42,7 +43,9 @@ class MarkdownExporter(TemplateExporter):
                                           'text/plain'
                                           ]
             },
-
+            'HighlightMagicsPreprocessor': {
+                'enabled':True
+                },
         })
-        c.merge(super(MarkdownExporter, self).default_config)
+        c.merge(super().default_config)
         return c
